@@ -36,15 +36,16 @@ class WorkflowEntity(Base):
     id = Column(String, primary_key=True)
     name = Column(String)
     description = Column(String, default='')
-    repo_uri = Column(String, default='')
+    git = Column(String, default='')
     version = Column(String, default='')
     username = Column(String, default='')
-    documentation_uri = Column(Text, default='')
     inputs = Column(Text, default='')
     parameters = Column(Text, default='')
     final_output = Column(Text, default='')
+    apps = Column(Text, default='')
     public = Column(Boolean, default=False)
     enable = Column(Boolean, default=True)
+    test = Column(Boolean, default=False)
     created = Column(DateTime, default=datetime.datetime.now)
     modified = Column(DateTime, default=datetime.datetime.now)
 
@@ -57,12 +58,15 @@ class AppEntity(Base):
     id = Column(String, primary_key=True)
     name = Column(String, default='')
     description = Column(String, default='')
-    repo_uri = Column(String, default='')
+    git = Column(String, default='')
     version = Column(String, default='')
     username = Column(String, default='')
-    definition = Column(Text, default='')
+    implementation = Column(Text, default='')
     inputs = Column(Text, default='')
     parameters = Column(Text, default='')
+    pre_exec = Column(Text, default='')
+    exec_methods = Column(Text, default='')
+    post_exec = Column(Text, default='')
     public = Column(Boolean, default=False)
 
 
@@ -87,10 +91,12 @@ class StepEntity(Base):
     workflow_id = Column(String)
     app_id = Column(String)
     map_uri = Column(String, default='')
+    map_glob = Column(String, default='*')
     map_regex = Column(String, default='')
     template = Column(Text, default='')
     exec_context = Column(String, default='local')
     exec_method = Column(String, default='auto')
+    exec_parameters = Column(Text, default='')
 
 
 class JobEntity(Base):
@@ -115,6 +121,7 @@ class JobEntity(Base):
     final_output = Column(Text, default='')
     exec_context = Column(Text, default='')
     exec_method = Column(Text, default='')
+    exec_parameters = Column(Text, default='')
     notifications = Column(Text, default='[]')
 
 
@@ -269,6 +276,7 @@ class DataSource:
                 JobEntity.final_output,
                 JobEntity.exec_context,
                 JobEntity.exec_method,
+                JobEntity.exec_parameters,
                 JobEntity.notifications
             ).\
                 filter(JobEntity.id == job_id).\
@@ -290,9 +298,10 @@ class DataSource:
                     'final_output': json.loads(row[10]),
                     'execution': {
                         'context': json.loads(row[11]),
-                        'method': json.loads(row[12])
+                        'method': json.loads(row[12]),
+                        'parameters': json.loads(row[13])
                     },
-                    'notifications': json.loads(row[13])
+                    'notifications': json.loads(row[14])
                 } for row in result
             ]
 
@@ -357,10 +366,12 @@ class DataSource:
                 StepEntity.number,
                 StepEntity.letter,
                 StepEntity.map_uri,
+                StepEntity.map_glob,
                 StepEntity.map_regex,
                 StepEntity.template,
                 StepEntity.exec_context,
-                StepEntity.exec_method
+                StepEntity.exec_method,
+                StepEntity.exec_parameters
             ).\
                 filter(StepEntity.workflow_id == workflow_id).\
                 filter(StepEntity.app_id == AppEntity.id).\
@@ -376,12 +387,14 @@ class DataSource:
                     'letter': row[5],
                     'map': {
                         'uri': row[6],
-                        'regex': row[7],
+                        'glob': row[7],
+                        'regex': row[8],
                     },
-                    'template': json.loads(row[8]),
+                    'template': json.loads(row[9]),
                     'execution': {
-                        'context': row[9],
-                        'method': row[10]
+                        'context': row[10],
+                        'method': row[11],
+                        'parameters': json.loads(row[12])
                     },
                     'depend': []
                 } for row in result
@@ -424,14 +437,15 @@ class DataSource:
                 WorkflowEntity.name,
                 WorkflowEntity.description,
                 WorkflowEntity.username,
-                WorkflowEntity.repo_uri,
+                WorkflowEntity.git,
                 WorkflowEntity.version,
-                WorkflowEntity.documentation_uri,
                 WorkflowEntity.inputs,
                 WorkflowEntity.parameters,
                 WorkflowEntity.final_output,
+                WorkflowEntity.apps,
                 WorkflowEntity.public,
-                WorkflowEntity.enable
+                WorkflowEntity.enable,
+                WorkflowEntity.test
             ).\
                 filter(WorkflowEntity.id == workflow_id).\
                 all()
@@ -442,14 +456,15 @@ class DataSource:
                     'name': row[1],
                     'description': row[2],
                     'username': row[3],
-                    'repo_uri': row[4],
+                    'git': row[4],
                     'version': row[5],
-                    'documentation_uri': row[6],
-                    'inputs': json.loads(row[7]),
-                    'parameters': json.loads(row[8]),
-                    'final_output': json.loads(row[9]),
+                    'inputs': json.loads(row[6]),
+                    'parameters': json.loads(row[7]),
+                    'final_output': json.loads(row[8]),
+                    'apps': json.loads(row[9]),
                     'public': row[10],
                     'enable': row[11],
+                    'test': row[12],
                     'steps': {}
                 } for row in result
             ]
@@ -495,13 +510,16 @@ class DataSource:
                 AppEntity.id,
                 AppEntity.name,
                 AppEntity.description,
-                AppEntity.repo_uri,
+                AppEntity.git,
                 AppEntity.version,
                 AppEntity.public,
                 AppEntity.username,
                 AppEntity.inputs,
                 AppEntity.parameters,
-                AppEntity.definition
+                AppEntity.implementation,
+                AppEntity.pre_exec,
+                AppEntity.exec_methods,
+                AppEntity.post_exec
             ).\
                 filter(StepEntity.workflow_id == workflow_id).\
                 filter(StepEntity.app_id == AppEntity.id).\
@@ -512,13 +530,16 @@ class DataSource:
                     'app_id': row[0],
                     'name': row[1],
                     'description': row[2],
-                    'repo_uri': row[3],
+                    'git': row[3],
                     'version': row[4],
                     'public': row[5],
                     'username': row[6],
                     'inputs': json.loads(row[7]),
                     'parameters': json.loads(row[8]),
-                    'definition': json.loads(row[9])
+                    'implementation': json.loads(row[9]),
+                    'pre_exec': json.loads(row[10]),
+                    'exec_methods': json.loads(row[11]),
+                    'post_exec': json.loads(row[12])
                 } for row in result
             }
 
@@ -618,14 +639,15 @@ class DataSource:
                 name=data['name'],
                 description=data['description'],
                 username=data['username'],
-                repo_uri=data['repo_uri'],
+                git=data['git'],
                 version=data['version'],
-                documentation_uri=data['documentation_uri'],
                 inputs=data['inputs'],
                 parameters=data['parameters'],
                 final_output=data['final_output'],
+                apps=data['apps'],
                 public=data['public'],
                 enable=data['enable'],
+                test=data['test'],
                 created=None,
                 modified=None
             ))
@@ -820,13 +842,17 @@ class DataSource:
                 id=app_id,
                 name=data['name'],
                 description=data['description'],
-                repo_uri=data['repo_uri'],
+                git=data['git'],
                 version=data['version'],
                 username=data['username'],
                 public=data['public'],
-                definition=data['definition'],
+                implementation=data['implementation'],
                 inputs=data['inputs'],
-                parameters=data['parameters']
+                parameters=data['parameters'],
+                pre_exec=data['pre_exec'],
+                exec_methods=data['exec_methods'],
+                post_exec=data['post_exec']
+
             ))
         except SQLAlchemyError as err:
             Log.an().error('sql exception [%s]', str(err))
@@ -1065,7 +1091,7 @@ class DataSource:
             data: a dictionary with the current keys:
                   ['workflow_id', 'app_id', 'name', 'number',
                   'letter', 'map_uri', 'map_regex','template',
-                  'exec_context', 'exec_method']
+                  'exec_context', 'exec_method', 'exec_parameters']
 
         Returns:
             On success: step id of the added StepEntity object.
@@ -1082,10 +1108,12 @@ class DataSource:
                 number=data['number'],
                 letter=data['letter'],
                 map_uri=data['map_uri'],
+                map_glob=data['map_glob'],
                 map_regex=data['map_regex'],
                 template=data['template'],
                 exec_context=data['exec_context'],
-                exec_method=data['exec_method']
+                exec_method=data['exec_method'],
+                exec_parameters=data['exec_parameters']
             ))
         except SQLAlchemyError as err:
             Log.an().error('sql exception [%s]', str(err))
@@ -1260,7 +1288,7 @@ class DataSource:
             data: a dictionary with the following keys:
                   ['workflow_id', 'name', 'username', 'work_uri', 'no_output_hash',
                   'inputs', 'parameters', 'output_uri','final_output',
-                  'exec_context', 'exec_method']
+                  'exec_context', 'exec_method', 'exec_parameters']
         Returns:
             On success: id of the added job.
             On failure: False.
@@ -1281,6 +1309,7 @@ class DataSource:
                 final_output=data['final_output'],
                 exec_context=data['exec_context'],
                 exec_method=data['exec_method'],
+                exec_parameters=data['exec_parameters'],
                 notifications=data['notifications']
             ))
         except SQLAlchemyError as err:
@@ -1653,15 +1682,18 @@ class DataSource:
                 return False
 
             app_id = self.add_app({
-                'name'          : valid_def['name'],
-                'description'   : valid_def['description'],
-                'repo_uri'      : valid_def['repo_uri'],
-                'version'       : valid_def['version'],
-                'username'      : valid_def['username'],
-                'public'        : valid_def['public'],
-                'definition'    : json.dumps(valid_def['definition']),
-                'inputs'        : json.dumps(valid_def['inputs']),
-                'parameters'    : json.dumps(valid_def['parameters'])
+                'name': valid_def['name'],
+                'description': valid_def['description'],
+                'git': valid_def['git'],
+                'version': valid_def['version'],
+                'username': valid_def['username'],
+                'public': valid_def['public'],
+                'implementation': json.dumps(valid_def['implementation']),
+                'inputs': json.dumps(valid_def['inputs']),
+                'parameters': json.dumps(valid_def['parameters']),
+                'pre_exec': json.dumps(valid_def['pre_exec']),
+                'exec_methods': json.dumps(valid_def['exec_methods']),
+                'post_exec': json.dumps(valid_def['post_exec'])
             })
             if not app_id:
                 Log.an().error(
@@ -1767,15 +1799,18 @@ class DataSource:
         if not self.update_app(
                 valid_def['app_id'],
                 {
-                    'name'          : valid_def['name'],
-                    'description'   : valid_def['description'],
-                    'repo_uri'      : valid_def['repo_uri'],
-                    'version'       : valid_def['version'],
-                    'username'      : valid_def['username'],
-                    'public'        : valid_def['public'],
-                    'definition'    : json.dumps(valid_def['definition']),
-                    'inputs'        : json.dumps(valid_def['inputs']),
-                    'parameters'    : json.dumps(valid_def['parameters'])
+                    'name': valid_def['name'],
+                    'description': valid_def['description'],
+                    'git': valid_def['git'],
+                    'version': valid_def['version'],
+                    'username': valid_def['username'],
+                    'public': valid_def['public'],
+                    'implementation': json.dumps(valid_def['implementation']),
+                    'inputs': json.dumps(valid_def['inputs']),
+                    'parameters': json.dumps(valid_def['parameters']),
+                    'pre_exec': json.dumps(valid_def['pre_exec']),
+                    'exec_methods': json.dumps(valid_def['exec_methods']),
+                    'post_exec': json.dumps(valid_def['post_exec']),
                 }
         ):
             Log.an().error(
@@ -1824,7 +1859,7 @@ class DataSource:
 
     def add_linked_apps(self, workflow_dict, base_path):
         """
-        Add apps referenced relatively from workflow defincition.
+        Add apps referenced relatively from workflow definition.
 
         Update workflow_dict to include new app IDs.
 
@@ -1846,10 +1881,7 @@ class DataSource:
                     # app not yet loaded
 
                     # import app definition
-                    if not os.path.isabs(step['app']):
-                        app_path = os.path.join(base_path, step['app'])
-                    else:
-                        app_path = step['app']
+                    app_path = os.path.join(base_path, 'apps', step['app'], 'app.yaml')
 
                     apps = self.import_apps_from_def(app_path)
                     if not apps:
@@ -1988,10 +2020,12 @@ class DataSource:
                 'number': step['number'],
                 'letter': step['letter'],
                 'map_uri': step['map']['uri'],
+                'map_glob': step['map']['glob'],
                 'map_regex': step['map']['regex'],
                 'template': json.dumps(step['template']),
                 'exec_context': step['execution']['context'],
-                'exec_method': step['execution']['method']
+                'exec_method': step['execution']['method'],
+                'exec_parameters': json.dumps(step['execution']['parameters'])
             })
             if not step_id:
                 Log.an().error('cannot add workflow step: %s', step_name)
@@ -2042,10 +2076,12 @@ class DataSource:
                         'number': step['number'],
                         'letter': step['letter'],
                         'map_uri': step['map']['uri'],
+                        'map_glob': step['map']['glob'],
                         'map_regex': step['map']['regex'],
                         'template': json.dumps(step['template']),
                         'exec_context': step['execution']['context'],
-                        'exec_method': step['execution']['method']
+                        'exec_method': step['execution']['method'],
+                        'exec_parameters': json.dumps(step['execution']['parameters'])
                     }
             ):
                 Log.an().error(
@@ -2135,17 +2171,18 @@ class DataSource:
 
             # insert workflow record
             workflow_id = self.add_workflow({
-                'name'              : valid_def['name'],
-                'description'       : valid_def['description'],
-                'username'          : valid_def['username'],
-                'inputs'            : json.dumps(valid_def['inputs']),
-                'repo_uri'          : valid_def['repo_uri'],
-                'documentation_uri' : valid_def['documentation_uri'],
-                'parameters'        : json.dumps(valid_def['parameters']),
-                'final_output'      : json.dumps(valid_def['final_output']),
-                'public'            : valid_def['public'],
-                'enable'            : valid_def['enable'],
-                'version'           : valid_def['version']
+                'name': valid_def['name'],
+                'description': valid_def['description'],
+                'username': valid_def['username'],
+                'inputs': json.dumps(valid_def['inputs']),
+                'apps': json.dumps(valid_def['apps']),
+                'git': valid_def['git'],
+                'parameters': json.dumps(valid_def['parameters']),
+                'final_output': json.dumps(valid_def['final_output']),
+                'public': valid_def['public'],
+                'enable': valid_def['enable'],
+                'test': valid_def['test'],
+                'version': valid_def['version']
             })
             if not workflow_id:
                 Log.an().error(
@@ -2290,17 +2327,18 @@ class DataSource:
         if not self.update_workflow(
                 valid_def['workflow_id'],
                 {
-                    'name':              valid_def['name'],
-                    'description':       valid_def['description'],
-                    'username':          valid_def['username'],
-                    'repo_uri':          valid_def['repo_uri'],
-                    'documentation_uri': valid_def['documentation_uri'],
-                    'inputs':            json.dumps(valid_def['inputs']),
-                    'parameters':        json.dumps(valid_def['parameters']),
-                    'final_output':      json.dumps(valid_def['final_output']),
-                    'public':            valid_def['public'],
-                    'enable':            valid_def['enable'],
-                    'version':           valid_def['version']
+                    'name': valid_def['name'],
+                    'description': valid_def['description'],
+                    'username': valid_def['username'],
+                    'git': valid_def['git'],
+                    'inputs': json.dumps(valid_def['inputs']),
+                    'parameters': json.dumps(valid_def['parameters']),
+                    'final_output': json.dumps(valid_def['final_output']),
+                    'apps': json.dumps(valid_def['apps']),
+                    'public': valid_def['public'],
+                    'enable': valid_def['enable'],
+                    'test': valid_def['test'],
+                    'version': valid_def['version']
                 }
         ):
             Log.an().error(
@@ -2444,18 +2482,19 @@ class DataSource:
 
             # insert job record
             job_id = self.add_job({
-                'workflow_id'   : valid_def['workflow_id'],
-                'name'          : valid_def['name'],
-                'username'      : valid_def['username'],
-                'work_uri'      : json.dumps(valid_def['work_uri']),
+                'workflow_id': valid_def['workflow_id'],
+                'name': valid_def['name'],
+                'username': valid_def['username'],
+                'work_uri': json.dumps(valid_def['work_uri']),
                 'no_output_hash': valid_def['no_output_hash'],
-                'inputs'        : json.dumps(valid_def['inputs']),
-                'parameters'    : json.dumps(valid_def['parameters']),
-                'output_uri'    : valid_def['output_uri'],
-                'final_output'  : json.dumps(valid_def['final_output']),
-                'exec_context'  : json.dumps(valid_def['execution']['context']),
-                'exec_method'   : json.dumps(valid_def['execution']['method']),
-                'notifications' : json.dumps(valid_def['notifications'])
+                'inputs': json.dumps(valid_def['inputs']),
+                'parameters': json.dumps(valid_def['parameters']),
+                'output_uri': valid_def['output_uri'],
+                'final_output': json.dumps(valid_def['final_output']),
+                'exec_context': json.dumps(valid_def['execution']['context']),
+                'exec_method': json.dumps(valid_def['execution']['method']),
+                'exec_parameters': json.dumps(valid_def['execution']['parameters']),
+                'notifications': json.dumps(valid_def['notifications'])
             })
             if not job_id:
                 Log.an().error(
